Package: plssem
Type: Package
Title: Complex Partial Least Squares Structural Equation Modeling
Version: 0.1.3
Authors@R: 
    person(given = "Kjell", family = "Solem Slupphaug",
           email = "slupphaugkjell@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0009-0005-8324-2834"))
Maintainer: Kjell Solem Slupphaug <slupphaugkjell@gmail.com>
Description: 
    Estimate complex Structural Equation Models (SEMs) by fitting Partial
    Least Squares Structural Equation Modeling (PLS-SEM) and Partial Least
    Squares consistent Structural Equation Modeling (PLSc-SEM) specifications
    that handle categorical data, non-linear relations, and multilevel
    structures. The implementation follows Lohmöller (1989) for the classic PLS-SEM
    algorithm, Dijkstra and Henseler (2015) for consistent PLSc-SEM, Dijkstra et al.,
    (2014) for nonlinear PLSc-SEM, and Schuberth, Henseler, Dijkstra (2018)
    for ordinal PLS-SEM and PLSc-SEM. Additional extensions are under development.
    The MC-OrdPLSc algorithm, used to handle ordinal interaction models is detailed
    in Slupphaug et al., (2026).
    References:
    Lohmöller, J.-B. (1989, ISBN:9783790803002).
      "Latent Variable Path Modeling with Partial Least Squares."
    Dijkstra, T. K., & Henseler, J. (2015).
      <doi:10.1016/j.jmva.2015.06.002>.
      "Consistent partial least squares path modeling."
    Dijkstra, T. K., & Schermelleh-Engel, K. (2014).
      <doi:10.1016/j.csda.2014.07.008>.
      "Consistent partial least squares for nonlinear structural equation models."
    Schuberth, F., Henseler, J., & Dijkstra, T. K. (2018).
      <doi:10.1007/s11135-018-0767-9>.
      "Partial least squares path modeling using ordinal categorical indicators."
    Slupphaug, K. Mehmetoglu, M. & Mittner, M. (2026).
      <doi:10.31234/osf.io/fwzj6_v1>.
      "Consistent Estimates from Biased Estimators: Monte-Carlo Consistent Partial Least Squares for Latent Interaction Models with Ordinal Indicators."
License: GPL-3
Encoding: UTF-8
LazyData: true
Imports: methods, stats, modsem (>= 1.0.20), lme4, lavaan, stringr,
        matrixStats, Rfast, collapse, mvnfast, reformulas, future,
        future.apply, progressr, FNN, MASS
Depends: R (>= 4.1.0)
URL: https://github.com/kss2k/plssem, https://kss2k.github.io/plssem/
Suggests: knitr, rmarkdown, ggplot2, mice, mvtnorm, pkgload
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-07-03 21:50:13 UTC; kss
Author: Kjell Solem Slupphaug [aut, cre] (ORCID:
    <https://orcid.org/0009-0005-8324-2834>)
Repository: CRAN
Date/Publication: 2026-07-03 22:10:02 UTC
Built: R 4.6.0; ; 2026-07-04 00:06:13 UTC; unix
