% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_generics.R
\name{parameter_estimates,PlsModel-method}
\alias{parameter_estimates,PlsModel-method}
\title{Parameter estimates for \code{PlsModel} objects}
\usage{
\S4method{parameter_estimates}{PlsModel}(object, colon.pi = TRUE, label.renamed.prod = FALSE, ...)
}
\arguments{
\item{object}{A \code{PlsModel} object.}

\item{colon.pi}{Logical; replace product-indicator labels with colon
notation (\code{X:Z}).}

\item{label.renamed.prod}{Logical; retain renamed product labels when colon
expansion occurs.}

\item{...}{Currently unused.}
}
\value{
A \code{PlsSemParTable} data frame.
}
\description{
Parameter estimates for \code{PlsModel} objects
}
