% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_generics.R
\name{is_admissible}
\alias{is_admissible}
\alias{is_admissible,PlsModel-method}
\title{Check whether a fitted model has admissible parameter estimates}
\usage{
is_admissible(object)

\S4method{is_admissible}{PlsModel}(object)
}
\arguments{
\item{object}{A fitted \code{PlsModel} object.}
}
\value{
A single logical value.
}
\description{
Check whether a fitted model has admissible parameter estimates
}
