% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R
\name{isMC_PLS}
\alias{isMC_PLS}
\title{Check if object is a MC-PLS model}
\usage{
isMC_PLS(object)
}
\arguments{
\item{object}{A fitted model object.}
}
\value{
\code{TRUE}/\code{FALSE}.
}
\description{
Check if object is a MC-PLS model
}
