% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_generics.R
\name{parameter_estimates}
\alias{parameter_estimates}
\title{Generic accessor for model parameter estimates}
\usage{
parameter_estimates(object, ...)
}
\arguments{
\item{object}{A fitted model object.}

\item{...}{Additional arguments passed to methods.}
}
\value{
A parameter table describing the fitted model.
}
\description{
Generic accessor for model parameter estimates
}
