% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{pls_construct_scores}
\alias{pls_construct_scores}
\title{Construct latent variable scores}
\usage{
pls_construct_scores(object, ...)
}
\arguments{
\item{object}{A fitted \code{plssem} model.}

\item{...}{Passed to [pls_predict()].}
}
\value{
A \code{PlsSemMatrix} of predicted latent scores.
}
\description{
Convenience wrapper around [pls_predict()] returning only the predicted
latent scores matrix.
}
