% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_generics.R
\name{boot}
\alias{boot}
\alias{boot,PlsModel-method}
\title{Retrieve bootstrap coefficient matrix}
\usage{
boot(object)

\S4method{boot}{PlsModel}(object)
}
\arguments{
\item{object}{A fitted model object.}
}
\value{
A \code{PlsSemMatrix} of bootstrap replicate parameter vectors
  (rows = replicates, cols = parameters).
}
\description{
Retrieve bootstrap coefficient matrix
}
\examples{
library(modsem)
library(plssem)

m <- "
  X =~ x1 + x2 + x3
  Z =~ z1 + z2 + z3
  Y =~ y1 + y2 + y3
  Y ~ X + Z + X:Z
"

fit <- pls(m, oneInt, bootstrap = TRUE, boot.R = 50)
boot(fit)

}
