% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_generics.R
\name{is_mcpls}
\alias{is_mcpls}
\alias{is_mcpls,PlsModel-method}
\title{Check whether an object uses the MC-OrdPLSc estimator}
\usage{
is_mcpls(object)

\S4method{is_mcpls}{PlsModel}(object)
}
\arguments{
\item{object}{A fitted model object.}
}
\value{
\code{TRUE} or \code{FALSE}.
}
\description{
Check whether an object uses the MC-OrdPLSc estimator
}
