% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R
\name{print.plssem}
\alias{print.plssem}
\title{Print a \code{plssem} object}
\usage{
\method{print}{plssem}(x, ...)
}
\arguments{
\item{x}{An object of class \code{plssem}.}

\item{...}{Additional arguments for compatibility with the generic.}
}
\value{
The input object, invisibly.
}
\description{
Print a \code{plssem} object
}
