% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{print.PlsSemPredict}
\alias{print.PlsSemPredict}
\title{Print a \code{PlsSemPredict} object}
\usage{
\method{print}{PlsSemPredict}(x, ...)
}
\arguments{
\item{x}{A \code{PlsSemPredict} object.}

\item{...}{Additional arguments for compatibility with the generic.}
}
\value{
The input object, invisibly.
}
\description{
Print a \code{PlsSemPredict} object
}
