% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{titanic}
\alias{titanic}
\title{Titanic Passenger Survival Data Set.}
\format{
A data frame with 1309 rows and 12 variables:
}
\source{
https://www.kaggle.com/c/titanic/data
}
\description{
This dataset has been re-packaged for convenience from
https://github.com/paulhendricks/titanic

\describe{
\item{PassengerId}{Passenger ID}
\item{Survived}{Passenger Survival Indicator}
\item{Pclass}{Passenger Class}
\item{Name}{Name}
\item{Sex}{Sex}
\item{Age}{Age}
\item{SibSp}{Number of Siblings/Spouses Aboard}
\item{Parch}{Number of Parents/Children Aboard}
\item{Ticket}{Ticket Number}
\item{Fare}{Passenger Fare}
\item{Cabin}{Cabin}
\item{Embarked}{Port of Embarkation}
\item{Female}{Dummy variable for \code{Sex="female"}}
}
}
\examples{

fit <- pls("Survived ~ Age + Female + Age:Female",
           data = titanic, ordered = "Survived")
pls_predict(fit, benchmark = "acc")
}
