% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R
\name{summary.plssem}
\alias{summary.plssem}
\title{Summarize a fitted \code{plssem} model}
\usage{
\method{summary}{plssem}(object, fit = TRUE, ...)
}
\arguments{
\item{object}{An object of class \code{plssem}.}

\item{fit}{Should fit measures be calculated?}

\item{...}{Additional arguments passed to or from methods.}
}
\value{
A \code{SummaryPlsSem} object containing formatted parameter estimates.
}
\description{
Summarize a fitted \code{plssem} model
}
