% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_generics.R
\name{coef,PlsModel-method}
\alias{coef,PlsModel-method}
\alias{coefficients,PlsModel-method}
\title{Extract coefficients from a \code{PlsModel} model}
\usage{
\S4method{coef}{PlsModel}(object, ...)

\S4method{coefficients}{PlsModel}(object, ...)
}
\arguments{
\item{object}{A \code{PlsModel} object.}

\item{...}{Currently unused.}
}
\value{
A named \code{PlsSemVector} of parameter estimates.
}
\description{
Extract coefficients from a \code{PlsModel} model
}
