% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predict,PlsModel-method}
\alias{predict,PlsModel-method}
\title{Predict from a fitted \code{PlsModel} (alias for \code{\link{pls_predict}})}
\usage{
\S4method{predict}{PlsModel}(object, newdata = NULL, ...)
}
\arguments{
\item{object}{A fitted \code{PlsModel} object.}

\item{newdata}{Optional new data matrix/data frame.}

\item{...}{Further arguments passed to \code{\link{pls_predict}}.}
}
\value{
A \code{PlsSemPredict} object.
}
\description{
Predict from a fitted \code{PlsModel} (alias for \code{\link{pls_predict}})
}
