% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R
\name{parameter_estimates.plssem}
\alias{parameter_estimates.plssem}
\title{Parameter estimates for \code{plssem} objects}
\usage{
\method{parameter_estimates}{plssem}(object, colon.pi = TRUE, label.renamed.prod = FALSE, ...)
}
\arguments{
\item{object}{An object of class \code{plssem}.}

\item{colon.pi}{Logical; whether to replace labels for interaction terms with colon notation.}

\item{label.renamed.prod}{Logical; whether renamed product labels should be retained when colon expansion occurs.}

\item{...}{Additional arguments (not used).}
}
\value{
A parameter table (data frame) describing the fitted model.
}
\description{
Parameter estimates for \code{plssem} objects
}
