% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_generics.R
\name{print.SummaryPlsSem}
\alias{print.SummaryPlsSem}
\title{Print a \code{SummaryPlsSem} object}
\usage{
\method{print}{SummaryPlsSem}(x, ...)
}
\arguments{
\item{x}{A \code{SummaryPlsSem} object as returned by
\code{\link[=summary,PlsModel-method]{summary}()}.}

\item{...}{Additional arguments for compatibility with the generic.}
}
\value{
The input object, invisibly.
}
\description{
Print a \code{SummaryPlsSem} object
}
