% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_generics.R
\name{vcov,PlsModel-method}
\alias{vcov,PlsModel-method}
\title{Extract the variance-covariance matrix from a \code{PlsModel} model}
\usage{
\S4method{vcov}{PlsModel}(object, ...)
}
\arguments{
\item{object}{A \code{PlsModel} object.}

\item{...}{Currently unused.}
}
\value{
A \code{PlsSemMatrix} (bootstrap-based vcov, or \code{NULL}).
}
\description{
Extract the variance-covariance matrix from a \code{PlsModel} model
}
